<?php
/*

Psuedo-forth (RPN) interpreter
+ HLL macros (if)
+ Option to disable backward jumps, preventing any possible infinite loops
+ Option to limit maximum number of tokens, bounding execution

*/

class ForthException extends Exception {};

class ForthProgram {
	
	private $opcodes;
	private $functions;
	private $allowLoops;
	public $maxTokens;
	
	public $stack;
	
	public function ForthProgram($src = null) {
		$this->opcodes = array();
		$this->tokens = array();
		$this->stack = array();
		
		$this->allowLoops = false;
		$this->maxTokens = 4000;
		
		$this->functions = array(
			// Order of arguments: stack-topmost first
			'.' => array(1, function($a) {
				return array();
			}),
			'*' => array(2, function($a, $b) {
				return array($a + $b);
			}),
			'+' => array(2, function($a, $b) {
				return array($a + $b);
			}),
			'-' => array(2, function($a, $b) {
				return array($a - $b);
			}),
			'/' => array(2, function($a, $b) {
				return array($b ? $a / $b : 0);
			}),
			'^' => array(2, function($a, $b) {
				return array(pow($a, $b));
			}),
			'dup' => array(1, function($last) {
				return array($last, $last);
			}),
			'swap' => array(2, function($a, $b) {
				return array($b, $a);
			}),
			'>' => array(2, function($a, $b) {
				return array(($a > $b) ? 1 : 0);
			}),
			'<' => array(2, function($a, $b) {
				return array(($a < $b) ? 1 : 0);
			}),
			'=' => array(2, function($a, $b) {
				return array(($a == $b) ? 1 : 0);
			}),
			'>=' => array(2, function($a, $b) {
				return array(($a >= $b) ? 1 : 0);
			}),
			'<=' => array(2, function($a, $b) {
				return array(($a <= $b) ? 1 : 0);
			}),
			'jmp' => array(1, function($rel) {
				return intval($rel); // jmp
			}),
			'jnz' => array(2, function($val, $rel) {
				return ($val != 0) ? intval($rel) : 0; // jmp
			}),
			'jz' => array(2, function($val, $rel) {
				return ($val == 0) ? intval($rel) : 0; // jmp
			}),
			'jgz' => array(2, function($val, $rel) {
				return ($val > 0) ? intval($rel) : 0; // jmp
			}),
		);
		
		if (! is_null($src)) $this->load($src);
	}
		
	public function load($src) {
		$this->opcodes = array();
		$this->tokens = $this->tokenise($src);
		
		if (is_int($this->maxTokens) && count($this->tokens) > $this->maxTokens) {
			throw new ForthException("Maximum number of tokens exceeded");
		}
		
		list($tkcount, $this->opcodes) = $this->parse(0);
		if ($tkcount != count($this->tokens)) {
			throw new ForthException('Mismatched braces in source');
		}
		
		$this->tokens = array(); // clear
	}
	
	private function tokenise($src) {
		if (! strlen(trim($src))) return array();
		return preg_split('/\s+/', $src);
	}
		
	private function parse($start = 0, $depth = 0) {
		$ops = array();
		
		for ($i = $start; $i < count($this->tokens); $i++) {
			$token = &$this->tokens[$i];
		
			if (! strlen($token)) {
				continue;
		
			} else if (is_numeric($token)) {
				$ops[] = floatval($token);
			
			} else if (array_key_exists(strtolower($token), $this->functions)) {
				$ops[] = strtolower($token);
				
			} else if ($token == '(') {
				list($i, $newops) = $this->parse($i + 1, $depth + 1);
				$ops = array_merge($ops, $newops);
			
			} else if ($token == ')') {
				break;
				
			} else if ($token == 'if') {
				//echo json_encode($this->tokens)."\n";
				
				if ($this->tokens[$i+1] != '(') {
					throw new ForthException('Syntax error');
				};
				list($ic, $condops) = $this->parse($i + 2, $depth + 1);
				if ($this->tokens[$ic+1] != '(') {
					throw new ForthException('Syntax error');
				};
				list($ie, $ifops) = $this->parse($ic + 2, $depth + 1);
				
				//
				
				$ops[] = count($ifops);
				$ops = array_merge($ops, $condops);				
				$ops[] = 'jz';
				$ops = array_merge($ops, $ifops);
				
				$i = $ie;
				
			} else {
				throw new ForthException('Unrecognised token "'.$token.'" in source');
			}
		}
		
		return array($i, $ops);
	}
	
	public function execute() {
	
		for($pc = 0; $pc < count($this->opcodes); $pc++) {
			$token = $this->opcodes[$pc];
		
			if (is_numeric($token)) {
				$this->stack[] = $token;
			
			} else if (array_key_exists($token, $this->functions)) {
				$popct = $this->functions[$token][0];
				if (count($this->stack) < $popct) {
					throw new ForthException('Stack underflow');
				}
				$args = array();
				for ($i = 0; $i < $popct; $i++) {
					$args[] = array_pop($this->stack);
				}
				$ret = call_user_func_array($this->functions[$token][1], $args);
				if (is_array($ret)) {
					$this->stack = array_merge($this->stack, $ret);
				} else if (is_int($ret)) {
					if ($this->allowLoops) {
						$pc += $ret;
					} else {
						$pc += max(0, $ret);
					}
				}
				
			} else {
				throw new ForthException('Corrupted stack');
			}
		}
	}
}

function ForthTest($src, $expect) {
	$f = new ForthProgram($src);

	$f->execute();
	echo str_pad(json_encode($src), 40)." => ".str_pad(json_encode($f->stack), 15);
	if ($f->stack == $expect) {
		echo " PASSED\n";
	} else {
		echo " FAILED! (expected ".json_encode($expect).")\n";
	}
}

function ForthTests() {
		
	ForthTest("", array());
	ForthTest("1 2 3", array(1,2,3));
	ForthTest("3 4 +", array(7));
	ForthTest("3 4 + dup +", array(14));
	ForthTest("1 .", array());
	ForthTest("2 4 ^", array(16));
	ForthTest("1 2 /", array(2));
	ForthTest("2 1 /", array(1/2));
	ForthTest("5 1 >", array(0));
	ForthTest("1 5 >", array(1));
	ForthTest("5 1 <", array(1));
	ForthTest("3 1 jz 101 1 jmp 102 103", array(101, 103));
	ForthTest("3 0 jz 101 1 jmp 102 103", array(102, 103));

	// Infinite loop protect
	ForthTest("-2 jmp", array());

	// Bracket parsing
	ForthTest("3 ( 2 )", array(3,2));

	// Basic conditionals
	ForthTest("1 if ( 0 ) ( 3 ) 4", array(1, 4));
	ForthTest("1 if ( 1 ) ( 3 ) 4", array(1, 3, 4));
	ForthTest("1 if ( 1 1 - ) ( 3 ) 4", array(1, 4));

	ForthTest("1 if ( 2 ) ( 3 if ( 0 ) ( 4 ) 5 ) 6", array(1, 3, 5, 6));
	
}

ForthTests();

